/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.remote.group;

import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jgroups.Channel;
import org.jgroups.ChannelException;
import org.jgroups.ExtendedReceiverAdapter;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.View;

import net.jini.config.Configuration;
import net.jini.config.ConfigurationException;

/**
 * <p>Group channel implementation using the JGroups framework.</p>
 * 
 * <p>Configured in section <code>net.scatterspace.group.jgroups</code>:<ul>
 *  <li><code>channelConfig</code>: a {@link File} passed to JGroups to
 *      configure the channel stack</li>
 *  <li><code>channelName</code>: the name used to connect to the channel, must
 *      be identical across all group members</li>
 * </ul></p>
 * 
 * <p>Logs to <code>net.scatterspace.group.jgroups</code>. Note that JGroups
 * has its own logging configuration via Commons Logging, the above logger
 * name is purely for this adapter.</p>
 */
public class JGroupsAdapter implements GroupChannel
{
    private Logger log_ = Logger.getLogger("net.scatterspace.group.jgroups");
    
    private String channelName_;
    private GroupNode node_;
    private Channel channel_;
    
    // Group state variables
    private int memberCount_ = 0;
    private Object memberCountMonitor_ = new Object();
    
    /**
     * Sets up and connects to the JGroups channel.
     * 
     * @param config Configuration to load from
     * @throws ConfigurationException Thrown if config options aren't specified
     * @throws ChannelException If there's an error constructing or connecting
     * to the JGroups channel
     */
    public JGroupsAdapter(Configuration config)
        throws ConfigurationException, GroupChannelException
    {
        // Used to configure the JGroups channel
        File channelConfig = (File) config.getEntry(
              "net.scatterspace.group.jgroups", "channelConfig", File.class);
        
        log_.log(Level.CONFIG,
                 "Using JGroups configuration file " +channelConfig);
        
        try
        {
            channel_ = new JChannel(channelConfig);
        }
        catch (ChannelException e)
        {
            throw new GroupChannelException("Error constructing JChannel", e);
        }
        
        // Used to ensure virtual synchrony
        channel_.setOpt(Channel.BLOCK, true);
        
        // Connect to a specific channel name
        channelName_ = (String) config.getEntry(
              "net.scatterspace.group.jgroups", "channelName", String.class);
    }
    
    public void connect() throws GroupChannelException
    {
        log_.log(Level.CONFIG, "Connecting to JGroups channel " + channelName_);
        
        // Adapter to the JGroups message passing
        channel_.setReceiver(new ChannelReceiver());
        
        try
        {
            channel_.connect(channelName_);
        }
        catch (ChannelException e)
        {
            throw new GroupChannelException(
                  "Error connecting JChannel to " + channelName_, e);
        }
    }
    
    private class ChannelReceiver extends ExtendedReceiverAdapter
    {
        @Override
        public void block()
        {
            log_.log(Level.FINEST, "Switched to blocking mode");
        }

        @Override
        public void unblock()
        {
            log_.log(Level.FINEST, "Blocking mode ceased");
        }

        @Override
        public void receive(Message msg)
        {
            log_.log(Level.FINER, "Received message: " + msg);
            
            // Currently we only know about messages that contain group ops
            if (!(msg.getObject() instanceof GroupOperation))
            {
                log_.log(Level.WARNING,
                         "Unknown type of message received: " + msg);
                return;
            }
            
            GroupOperation op = (GroupOperation) msg.getObject();
            op.execute(node_);
        }

        @Override
        public void viewAccepted(View newView)
        {
            log_.log(Level.FINE, "New view has taken effect: " + newView);
            
            synchronized (memberCountMonitor_)
            {
                memberCount_ = newView.getMembers().size();
                memberCountMonitor_.notify();
            }
        }
    }
    
    public void setNode(GroupNode node) throws IllegalStateException
    {
        if (node_ != null)
            throw new IllegalStateException("Node already set");
        
        node_ = node;
    }
    
    public void destroy()
    {
        log_.log(Level.FINE, "Destroying the channel");
        
        channel_.close();
    }

    /**
     * Package protected for testing.  Gets the number of members currently
     * known in the group.
     * @return Number of members
     */
    int getMemberCount()
    {
        synchronized (memberCountMonitor_)
        {
            return memberCount_;
        }
    }

    /**
     * Package protected for testing.  Monitor object to hear when the member
     * count changes.
     * @return Object to call {@link #wait()} on.
     */
    Object getMemberCountMonitor()
    {
        return memberCountMonitor_;
    }
}
