/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.client;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.jini.config.Configuration;
import net.jini.config.ConfigurationException;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceTemplate;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lookup.LookupCache;
import net.jini.lookup.ServiceDiscoveryEvent;
import net.jini.lookup.ServiceDiscoveryListener;
import net.jini.lookup.ServiceDiscoveryManager;
import net.jini.lookup.ServiceItemFilter;

/**
 * Utility for both registrar and service discovery for clients, using a
 * {@link Configuration} or passed discovery parameters.
 * It reads the following configuration entries from the
 * <code>net.scatterspace.client</code> component:
 * 
 * <ul>
 *  <li><b>discoveryGroups</b> (<code>String[]</code>),
 *      defines the set of multicast discovery groups that the finder will use.
 *      Default value is {@link DiscoveryGroupManagement#ALL_GROUPS}.
 *      See also {@link LookupDiscoveryManager}.</li>
 *  <li><b>lookupLocators</b> (<code>{@link LookupLocator}[]</code>),
 *      defines the set of unicast lookup locators that will be used for
 *      discovery.  Default value is <code>null</code> (no unicast lookup
 *      discovery).</li>
 * </ul>
 * 
 * <p>This utility makes use of both {@link LookupDiscoveryManager} and
 * {@link ServiceDiscoveryManager} and will pass through the supplied
 * <code>Configuration</code> object, allowing them to be configured also.</p>
 * 
 * <p>It logs to <code>net.scatterspace.client.ServiceFinder</code>.</p>
 * 
 * @see LookupDiscoveryManager
 * @see ServiceDiscoveryManager
 */
public class ServiceFinder
{
    private static Logger log__ =
        Logger.getLogger(ServiceFinder.class.getName());
    
    private LookupDiscoveryManager lookupManager_;
    private ServiceDiscoveryManager serviceManager_;
    
    /** Maximum number of services to consider in lookups */
    private int maximumServices_ = Integer.MAX_VALUE;
    
    /**
     * Instantiate the service finder and begin performing the discovery
     * process.
     * 
     * @param config Configuration for this class, also passed through to
     * LookupDiscoveryManager and ServiceDiscoveryManager.
     * @throws ConfigurationException if an entry cannot be read from the
     * supplied <code>config</code> object 
     * @throws IOException may be thrown by the discovery process
     */
    public ServiceFinder(Configuration config)
        throws ConfigurationException, IOException
    {
        String[] groups = (String[]) config.getEntry(
                "net.scatterspace.client", "discoveryGroups",
                String[].class, DiscoveryGroupManagement.ALL_GROUPS
        );
        
        LookupLocator[] locators = (LookupLocator[]) config.getEntry(
                "net.scatterspace.client", "lookupLocators",
                LookupLocator[].class, null
        );
        
        log__.log(Level.CONFIG, "Configured discovery groups are " +
                  Arrays.toString(groups) + " and unicast locators " +
                  Arrays.toString(locators));
        
        lookupManager_ =
            new LookupDiscoveryManager(groups, locators, null, config);
        
        serviceManager_ =
            new ServiceDiscoveryManager(lookupManager_, null, config);
    }
    
    /**
     * Instantiate the service finder and begin performing the discovery
     * process.  Instead of a config, the discovery groups and locators can be
     * specified.
     * 
     * @param groups Discovery group names passed through to
     * LookupDiscoveryManager
     * @param locators Discovery lookup locators passed through to
     * LookupDiscoveryManager 
     * @throws IOException may be thrown by the discovery process
     */
    public ServiceFinder(String[] groups, LookupLocator[] locators)
        throws IOException
    {
        log__.log(Level.CONFIG, "Configured discovery groups are " +
                Arrays.toString(groups) + " and unicast locators " +
                Arrays.toString(locators));
        
        lookupManager_ = new LookupDiscoveryManager(groups, locators, null);
        serviceManager_ = new ServiceDiscoveryManager(lookupManager_, null);
    }
    
    /**
     * Perform a lookup for a particular service that matches one of the
     * supplied service types.
     * 
     * @param serviceTypes Classes of possible matching services
     * @return Service result container that will remain valid
     * @see ServiceDiscoveryManager#lookup(ServiceTemplate, ServiceItemFilter, long)
     */
    public ServiceFinderResult lookup(Class<?>[] serviceTypes)
        throws RemoteException, InterruptedException
    {
        return lookup(new ServiceTemplate(null, serviceTypes, null), null);
    }
    
    /**
     * Perform a lookup for a particular service that matches the supplied
     * template and filter.z
     * 
     * @param template Service matching template
     * @param filter Service matching filter
     * @return Service result container that will remain valid
     * @see ServiceDiscoveryManager#lookup(ServiceTemplate, ServiceItemFilter, long)
     */
    public ServiceFinderResult lookup(ServiceTemplate template,
                                      ServiceItemFilter filter)
        throws RemoteException
    {
        log__.log(Level.FINE, "Performing lookup with template " + template +
                  ", filter " + filter);
        
        return new ResultImpl(template, filter);
    }
    
    /** Real implementation for found results */
    private class ResultImpl implements ServiceFinderResult
    {
        private ServiceTemplate template_;
        private ServiceItemFilter filter_;
        private LookupCache cache_;
        
        /** Use the object monitor to know about arrivals in the cache */
        private Object serviceMonitor_ = new Object();
        
        /** Stores previously returned services so they're avoided later */
        private Set<ServiceID> previousServices = new HashSet<ServiceID>();
        
        public ResultImpl(ServiceTemplate template, ServiceItemFilter filter)
            throws RemoteException
        {
            template_ = template;
            filter_ = filter;
            
            cache_ = serviceManager_.createLookupCache(template, filter,
                                                       new CacheListener());
        }
        
        public Object getService(long waitDuration)
            throws RemoteException, InterruptedException
        {
            long start, now;
            start = now = System.currentTimeMillis();
            
            ServiceItem[] results = new ServiceItem[0];
            synchronized (serviceMonitor_)
            {
                results = cache_.lookup(null, maximumServices_);
                
                if (log__.isLoggable(Level.FINER))
                    log__.log(Level.FINER, "Retrieved " + results.length +
                              " from lookup cache on first attempt");
                
                if (results == null || results.length == 0)
                    serviceManager_.lookup(template_, filter_, waitDuration);
                
                while ((results == null || results.length == 0) &&
                        (waitDuration <= 0 || now < start + waitDuration))
                {
                    if (log__.isLoggable(Level.FINEST))
                        log__.log(Level.FINEST, "Waiting on lookup event for " +
                                  waitDuration + " ms");
                    
                    serviceMonitor_.wait(waitDuration);
                    results = cache_.lookup(null, maximumServices_);
                }
                
                if (log__.isLoggable(Level.FINER))
                    log__.log(Level.FINER, "Retrieved " + results.length +
                              " from lookup cache");
            }
            
            if (log__.isLoggable(Level.FINEST))
                log__.log(Level.FINEST, "Services retrieved from cache: " +
                          Arrays.toString(results));
            
            ServiceItem chosen = null;
            if (results != null)
            {
                for (ServiceItem service : results)
                {
                    if (!previousServices.contains(service.serviceID))
                    {
                        previousServices.add(service.serviceID);
                        chosen = service;
                        break;
                    }
                }
            }
            
            if (chosen == null)
            {
                if (log__.isLoggable(Level.FINER))
                    log__.log(Level.FINER, "Unable to find an unseen service," +
                              " returning the first available");
                if (results != null && results.length > 0)
                    chosen = results[0];
            }
            
            log__.log(Level.FINER, "Returning service " + chosen);
            return (chosen != null) ? chosen.service : null;
        }
        
        private class CacheListener implements ServiceDiscoveryListener
        {
            public void serviceAdded(ServiceDiscoveryEvent event)
            {
                synchronized (serviceMonitor_)
                {
                    serviceMonitor_.notifyAll();
                }
            }

            public void serviceChanged(ServiceDiscoveryEvent arg0)
            {
                // Doesn't affect us
            }

            public void serviceRemoved(ServiceDiscoveryEvent arg0)
            {
                // Doesn't affect us
            }
        }
    }
}
