/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.jmx;

import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.jini.config.Configuration;
import net.jini.config.ConfigurationException;
import net.jini.config.ConfigurationFile;
import net.jini.config.ConfigurationProvider;
import net.jini.loader.pref.PreferredClassLoader;

/**
 * JVM agent that sets up a JMX server over Jini.  Takes one argument, which is
 * passed to {@link ConfigurationProvider} and is normally a path to a file that
 * is used by {@link ConfigurationFile}.
 */
public class JiniAgent
{
    private static Logger log_ = Logger.getLogger(JiniAgent.class.getName());
    
    /** May not be constructed */
    private JiniAgent() { }
    
    /**
     * @param agentArgs Configuration file path
     */
    public static void premain(String agentArgs)
        throws ConfigurationException, IOException
    {
        Configuration config = ConfigurationProvider.getInstance(
                new String[] { agentArgs });
        
        String codebase = (String) config.getEntry(
                "net.scatterspace.jmx", "codebase", String.class, null);
        if (codebase == null)
        {
            log_.log(Level.CONFIG, "Loading JMX-Jini agent without a codebase");
            new MBeanAdapterImpl(config);
        }
        else
        {
            URL[] classpath = (URL[]) config.getEntry("net.scatterspace.jmx",
                    "classpath", URL[].class, new URL[0]);
            
            log_.log(Level.CONFIG, "Loading JMX-Jini agent with classpath " +
                     Arrays.toString(classpath) + " and codebase " + codebase);
            
            ClassLoader oldCl = Thread.currentThread().getContextClassLoader();
            
            ClassLoader cl = new PreferredClassLoader(classpath,
                    oldCl, codebase, false);
            Thread.currentThread().setContextClassLoader(cl);
            
            try
            {
                Class<?> implClazz = Class.forName(
                        "net.scatterspace.jmx.MBeanAdapterImpl", false, cl);
                implClazz.getConstructor(Configuration.class).newInstance(config);
            }
            catch (Exception e)
            {
                log_.log(Level.SEVERE,
                         "Unable to load JMX-Jini agent with codebase", e);
            }
            finally
            {
                Thread.currentThread().setContextClassLoader(oldCl);
            }
        }
    }
    
    public static void agentmain(String agentArgs)
        throws ConfigurationException, IOException
    {
        premain(agentArgs);
    }
}
