/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.jmx;

import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServerConnection;
import javax.management.NotCompliantMBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;

import com.sun.jini.lookup.entry.BasicServiceType;
import com.sun.jini.start.LifeCycle;

import net.jini.config.Configuration;
import net.jini.config.ConfigurationException;
import net.jini.config.ConfigurationProvider;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.LookupDiscovery;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.export.Exporter;
import net.jini.id.Uuid;
import net.jini.id.UuidFactory;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;
import net.jini.lookup.JoinManager;
import net.jini.lookup.entry.ServiceInfo;

/**
 * Implementation of the remote MBean connector which exports itself and acts as
 * an adapter to the local JVM's MBean server (via
 * {@link ManagementFactory#getPlatformMBeanServer()}.
 */
public class MBeanAdapterImpl implements MBeanConnector
{
    private static final String MODULE = "net.scatterspace.jmx";
    
    private Logger log_ = Logger.getLogger(getClass().getName());
    private MBeanServerConnection conn_;
    private MBeanConnector proxy_;
    
    /**
     * Constructor designed for the {@link com.sun.jini.start} service starter.
     * Exports the service and prepares it for use.
     * @param args Passed to {@link ConfigurationProvider}, path to config file 
     */
    public MBeanAdapterImpl(String[] args, LifeCycle lifecycle)
        throws IOException, ConfigurationException
    {
        this(ConfigurationProvider.getInstance(args));
    }
    
    /**
     * Load and export the service.
     */
    public MBeanAdapterImpl(Configuration config)
        throws IOException, ConfigurationException
    {
        // Prepare the config for later
        Exporter exporter = (Exporter)
            config.getEntry(MODULE, "serverExporter", Exporter.class,
                            new BasicJeriExporter(
                                  TcpServerEndpoint.getInstance(0),
                                  new BasicILFactory()));
        
        String[] groups = (String[])
            config.getEntry(MODULE, "initialMemberGroups",
                            String[].class, LookupDiscovery.ALL_GROUPS);
        
        LookupLocator[] locs = (LookupLocator[])
            config.getEntry(MODULE, "initialLocators",
                            LookupLocator[].class, null);
        
        Collection<Entry> attrs = new ArrayList<Entry>();
        attrs.add(new ServiceInfo("JMX-Jini", "ScatterSpace",
                                  "ScatterSpace", "0.1", null, null));
        attrs.add(new BasicServiceType("JMX-Jini"));
        
        attrs.addAll(Arrays.asList((Entry[])
            config.getEntry(MODULE, "initialAttrs",
                            Entry[].class, new Entry[0])));
        
        log_.log(Level.FINE, "MBeanServer loading, exporting with " + exporter);
        
        // Use the local MBean server as the server to delegate to
        conn_ = ManagementFactory.getPlatformMBeanServer();
        log_.log(Level.FINE, "Using local MBean server " + conn_);
        
        // Now set up the Jini half of things so we're publicly visible
        proxy_ = (MBeanConnector) exporter.export(this);
        
        // Each service ought to have a unique id - this can be assigned
        // us by reggie but we'll generate one ourselves
        Uuid svcUuid = UuidFactory.generate();
        
        ServiceID svcId = new ServiceID(svcUuid.getMostSignificantBits(),
                                        svcUuid.getLeastSignificantBits());
        
        // Discover registrars and join them
        log_.log(Level.INFO,
                 "MBeanServer joining to groups " + Arrays.toString(groups) +
                 ", locators " + Arrays.toString(locs) +
                 " (ServiceID: " + svcId + ")");
        
        LookupDiscoveryManager ldm =
            new LookupDiscoveryManager(groups, locs, null, config);
        new JoinManager(proxy_, attrs.toArray(new Entry[0]), svcId,
                        ldm, null, config);
    }

    public void addNotificationListener(ObjectName name,
                                        NotificationConnector listener,
                                        NotificationFilter filter,
                                        Object handback)
            throws InstanceNotFoundException, IOException
    {
        conn_.addNotificationListener(name,
              new ServerNotificationConnector(listener), filter, handback);
    }

    public void addNotificationListener(ObjectName name, ObjectName listener,
                                        NotificationFilter filter,
                                        Object handback)
            throws InstanceNotFoundException, IOException
    {
        conn_.addNotificationListener(name, listener, filter, handback);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      Object[] params, String[] signature)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, IOException
    {
        return conn_.createMBean(className, name, params, signature);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      ObjectName loaderName, Object[] params,
                                      String[] signature)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, InstanceNotFoundException, IOException
    {
        return conn_
                .createMBean(className, name, loaderName, params, signature);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      ObjectName loaderName)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, InstanceNotFoundException, IOException
    {
        return conn_.createMBean(className, name, loaderName);
    }

    public ObjectInstance createMBean(String className, ObjectName name)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, IOException
    {
        return conn_.createMBean(className, name);
    }

    public Object getAttribute(ObjectName name, String attribute)
            throws MBeanException, AttributeNotFoundException,
            InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.getAttribute(name, attribute);
    }

    public AttributeList getAttributes(ObjectName name, String[] attributes)
            throws InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.getAttributes(name, attributes);
    }

    public String getDefaultDomain() throws IOException
    {
        return conn_.getDefaultDomain();
    }

    public String[] getDomains() throws IOException
    {
        return conn_.getDomains();
    }

    public Integer getMBeanCount() throws IOException
    {
        return conn_.getMBeanCount();
    }

    public MBeanInfo getMBeanInfo(ObjectName name)
            throws InstanceNotFoundException, IntrospectionException,
            ReflectionException, IOException
    {
        return conn_.getMBeanInfo(name);
    }

    public ObjectInstance getObjectInstance(ObjectName name)
            throws InstanceNotFoundException, IOException
    {
        return conn_.getObjectInstance(name);
    }

    public Object invoke(ObjectName name, String operationName,
                         Object[] params, String[] signature)
            throws InstanceNotFoundException, MBeanException,
            ReflectionException, IOException
    {
        return conn_.invoke(name, operationName, params, signature);
    }

    public boolean isInstanceOf(ObjectName name, String className)
            throws InstanceNotFoundException, IOException
    {
        return conn_.isInstanceOf(name, className);
    }

    public boolean isRegistered(ObjectName name) throws IOException
    {
        return conn_.isRegistered(name);
    }

    public Set<ObjectInstance> queryMBeans(ObjectName name, QueryExp query)
            throws IOException
    {
        return conn_.queryMBeans(name, query);
    }

    public Set<ObjectName> queryNames(ObjectName name, QueryExp query)
            throws IOException
    {
        return conn_.queryNames(name, query);
    }

    public void removeNotificationListener(ObjectName name,
                                           NotificationListener listener,
                                           NotificationFilter filter,
                                           Object handback)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener, filter, handback);
    }

    public void removeNotificationListener(ObjectName name,
                                           NotificationListener listener)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener);
    }

    public void removeNotificationListener(ObjectName name,
                                           ObjectName listener,
                                           NotificationFilter filter,
                                           Object handback)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener, filter, handback);
    }

    public void removeNotificationListener(ObjectName name, ObjectName listener)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener);
    }

    public void setAttribute(ObjectName name, Attribute attribute)
            throws InstanceNotFoundException, AttributeNotFoundException,
            InvalidAttributeValueException, MBeanException,
            ReflectionException, IOException
    {
        conn_.setAttribute(name, attribute);
    }

    public AttributeList setAttributes(ObjectName name, AttributeList attributes)
            throws InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.setAttributes(name, attributes);
    }

    public void unregisterMBean(ObjectName name)
            throws InstanceNotFoundException, MBeanRegistrationException,
            IOException
    {
        conn_.unregisterMBean(name);
    }
    
    /**
     * Adapter for a (remote) {@link NotificationConnector} to the
     * {@link NotificationListener} interface.
     */
    private class ServerNotificationConnector implements NotificationListener
    {
        private NotificationConnector connector_;
        
        public ServerNotificationConnector(NotificationConnector connector)
        {
            connector_ = connector;
        }
        
        @Override
        public void handleNotification(Notification notification,
                                       Object handback)
        {
            try
            {
                connector_.handleNotification(notification, handback);
            }
            catch (RemoteException e)
            {
                log_.log(Level.WARNING, "Unable to deliver notification", e);
            }
        }
    }
}
