/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.jmx;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.remote.JMXServiceURL;

import com.sun.jini.config.ConfigUtil;

import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.lookup.entry.ServiceInfo;

/**
 * <p>Utility to read a {@link JMXServiceURL} so it can be used for lookup of a
 * Jini-registered service.  Example URLs:<ul>
 *   <li>service:jmx:jini://all/servicetype:name=JavaSpace</li>
 *   <li>service:jmx:jini://all/servicetype:name=JavaSpace,vendor=MyVendor</li>
 *   <li>service:jmx:jini://mygroup/serviceid:1234567890,7894560123</li>
 * </ul></p>
 * <p>The host in the URL is the discovery group (<i>all</i> is a synonym for
 * {@link DiscoveryGroupManagement#ALL_GROUPS}).  Multiple groups may be given,
 * period-separated.</p>
 * <p>The path is separated by slashes, each being a single attribute entry
 * which can have multiple fields set (comma separated) with key=value pairs.
 * </p>
 */
public class SearchURLReader
{
    private Logger log_ = Logger.getLogger(getClass().getName());
    private String[] discoveryGroups_;
    private Collection<Entry> attributes_ = new ArrayList<Entry>();
    private ServiceID svcId_;
    
    public SearchURLReader(JMXServiceURL url)
    {
        log_.log(Level.FINE, "Parsing JMX URL " + url);
        
        if (url.getHost().equalsIgnoreCase("all"))
            discoveryGroups_ = DiscoveryGroupManagement.ALL_GROUPS;
        else
            discoveryGroups_ = url.getHost().split("\\.");
        
        log_.log(Level.FINER, "Discovery groups: " +
                 Arrays.toString(discoveryGroups_));
        
        String[] components = url.getURLPath().split("/");
        // ServiceInfo:name=a,vendor=blah
        for (String com : components)
        {
            String[] types = com.split(":", 2);
            Entry entry = null;
            
            if (types[0].equalsIgnoreCase("ServiceInfo"))
            {
                entry = new ServiceInfo();
                if (types.length > 1)
                    setAttributes(entry, types[1]);
            }
            else if (types[0].equalsIgnoreCase("ServiceID"))
            {
                if (types.length > 1)
                {
                    svcId_ = ConfigUtil.createServiceID(types[1]);
                    log_.log(Level.FINER, "Adding service ID: " + svcId_);
                }
            }
            else if (types[0].length() > 0)
            {
                log_.log(Level.WARNING, "Unknown attribute type " + types[0]);
            }
            
            if (entry != null)
            {
                log_.log(Level.FINER, "Added service attribute: " + entry);
                attributes_.add(entry);
            }
        }
    }
    
    /**
     * Sets attributes on the supplied entry from the attribute string.  In the
     * format of comma separated key=value pairs.  For example:<ul>
     *   <li>akey=avalue,bkey=bvalue,ckey=cvalue</li>
     * </ul>
     * @param entry Entry to set attributes on
     * @param attrInput Input string (comma separated)
     */
    private void setAttributes(Entry entry, String attrInput)
    {
        String[] attrs = attrInput.split(",");
        for (String attr : attrs)
        {
            String[] kv = attr.split("=", 2);
            if (kv.length < 2)
            {
                log_.log(Level.WARNING, "Attribute " + kv[0] +
                         " missing value");
            }
            else
            {
                try
                {
                    log_.log(Level.FINEST, "Setting " + kv[0] + " to " +
                             kv[1] + " on " + entry);
                    entry.getClass().getField(kv[0]).set(entry, kv[1]);
                }
                catch (NoSuchFieldException e)
                {
                    log_.log(Level.WARNING, "Attribute " + kv[0] +
                             " doesn't exist on " +
                             entry.getClass().getSimpleName());
                }
                catch (IllegalAccessException e)
                {
                    log_.log(Level.WARNING, "Unable to set " + kv[0], e);
                }
            }
        }
    }

    public String[] getDiscoveryGroups()
    {
        return discoveryGroups_;
    }

    public Entry[] getAttributes()
    {
        Entry[] ret = attributes_.toArray(new Entry[0]);
        return (ret.length > 0 ? ret : null);
    }

    public ServiceID getServiceId()
    {
        return svcId_;
    }
}
