/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.jmx.jini;

import java.io.IOException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServerConnection;
import javax.management.NotCompliantMBeanException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;

import net.jini.export.Exporter;
import net.jini.jeri.BasicILFactory;
import net.jini.jeri.BasicJeriExporter;
import net.jini.jeri.tcp.TcpServerEndpoint;
import net.scatterspace.jmx.MBeanConnector;
import net.scatterspace.jmx.NotificationConnector;

/**
 * <p>Adapter for a client to use that makes a (remote) {@link MBeanConnector}
 * service available through the {@link MBeanServerConnection} interface.</p>
 * 
 * <p>Also exports {@link NotificationListener}s in
 * {@link #addNotificationListener(ObjectName, NotificationListener, NotificationFilter, Object)}
 * using {@link NotificationConnector} so events can traverse the network. 
 */
public class ClientMBeanAdapter implements MBeanServerConnection
{
    private Logger log_ = Logger.getLogger(getClass().getName());
    private MBeanConnector conn_;
    
    public ClientMBeanAdapter(MBeanConnector conn)
    {
        conn_ = conn;
    }
    
    public void addNotificationListener(ObjectName name,
                                        final NotificationListener listener,
                                        NotificationFilter filter,
                                        Object handback)
            throws InstanceNotFoundException, IOException
    {
        log_.log(Level.FINER, "Adding notification listener: " + name);
        try
        {
            /** Adapter that will be exported and made remotely available
             * to the local listener object */
            NotificationConnector adapter = new NotificationConnector()
            {
                public void handleNotification(Notification notification,
                                               Object handback)
                {
                    listener.handleNotification(notification, handback);
                }
            };
            
            Exporter exp = new BasicJeriExporter(
                     TcpServerEndpoint.getInstance(0),
                     new BasicILFactory());
            NotificationConnector remoteListener =
                (NotificationConnector) exp.export(adapter);
            
            conn_.addNotificationListener(name, remoteListener,
                                          filter, handback);
        }
        catch (RuntimeException e)
        {
            log_.log(Level.WARNING, "Exception in addNotificationListener", e);
            throw e;
        }
        catch (InstanceNotFoundException e)
        {
            log_.log(Level.WARNING, "Exception in addNotificationListener", e);
            throw e;
        }
        catch (IOException e)
        {
            log_.log(Level.WARNING, "Exception in addNotificationListener", e);
            throw e;
        }
    }

    public void addNotificationListener(ObjectName name, ObjectName listener,
                                        NotificationFilter filter,
                                        Object handback)
            throws InstanceNotFoundException, IOException
    {
        conn_.addNotificationListener(name, listener, filter, handback);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      Object[] params, String[] signature)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, IOException
    {
        return conn_.createMBean(className, name, params, signature);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      ObjectName loaderName, Object[] params,
                                      String[] signature)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, InstanceNotFoundException, IOException
    {
        return conn_
                .createMBean(className, name, loaderName, params, signature);
    }

    public ObjectInstance createMBean(String className, ObjectName name,
                                      ObjectName loaderName)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, InstanceNotFoundException, IOException
    {
        return conn_.createMBean(className, name, loaderName);
    }

    public ObjectInstance createMBean(String className, ObjectName name)
            throws ReflectionException, InstanceAlreadyExistsException,
            MBeanRegistrationException, MBeanException,
            NotCompliantMBeanException, IOException
    {
        return conn_.createMBean(className, name);
    }

    public Object getAttribute(ObjectName name, String attribute)
            throws MBeanException, AttributeNotFoundException,
            InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.getAttribute(name, attribute);
    }

    public AttributeList getAttributes(ObjectName name, String[] attributes)
            throws InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.getAttributes(name, attributes);
    }

    public String getDefaultDomain() throws IOException
    {
        return conn_.getDefaultDomain();
    }

    public String[] getDomains() throws IOException
    {
        return conn_.getDomains();
    }

    public Integer getMBeanCount() throws IOException
    {
        return conn_.getMBeanCount();
    }

    public MBeanInfo getMBeanInfo(ObjectName name)
            throws InstanceNotFoundException, IntrospectionException,
            ReflectionException, IOException
    {
        return conn_.getMBeanInfo(name);
    }

    public ObjectInstance getObjectInstance(ObjectName name)
            throws InstanceNotFoundException, IOException
    {
        return conn_.getObjectInstance(name);
    }

    public Object invoke(ObjectName name, String operationName,
                         Object[] params, String[] signature)
            throws InstanceNotFoundException, MBeanException,
            ReflectionException, IOException
    {
        return conn_.invoke(name, operationName, params, signature);
    }

    public boolean isInstanceOf(ObjectName name, String className)
            throws InstanceNotFoundException, IOException
    {
        return conn_.isInstanceOf(name, className);
    }

    public boolean isRegistered(ObjectName name) throws IOException
    {
        return conn_.isRegistered(name);
    }

    public Set<ObjectInstance> queryMBeans(ObjectName name, QueryExp query)
            throws IOException
    {
        return conn_.queryMBeans(name, query);
    }

    public Set<ObjectName> queryNames(ObjectName name, QueryExp query)
            throws IOException
    {
        return conn_.queryNames(name, query);
    }

    public void removeNotificationListener(ObjectName name,
                                           NotificationListener listener,
                                           NotificationFilter filter,
                                           Object handback)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener, filter, handback);
    }

    public void removeNotificationListener(ObjectName name,
                                           NotificationListener listener)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener);
    }

    public void removeNotificationListener(ObjectName name,
                                           ObjectName listener,
                                           NotificationFilter filter,
                                           Object handback)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener, filter, handback);
    }

    public void removeNotificationListener(ObjectName name, ObjectName listener)
            throws InstanceNotFoundException, ListenerNotFoundException,
            IOException
    {
        conn_.removeNotificationListener(name, listener);
    }

    public void setAttribute(ObjectName name, Attribute attribute)
            throws InstanceNotFoundException, AttributeNotFoundException,
            InvalidAttributeValueException, MBeanException,
            ReflectionException, IOException
    {
        conn_.setAttribute(name, attribute);
    }

    public AttributeList setAttributes(ObjectName name, AttributeList attributes)
            throws InstanceNotFoundException, ReflectionException, IOException
    {
        return conn_.setAttributes(name, attributes);
    }

    public void unregisterMBean(ObjectName name)
            throws InstanceNotFoundException, MBeanRegistrationException,
            IOException
    {
        conn_.unregisterMBean(name);
    }
}
