/**
 * Copyright (c) 2008 CDO2 Limited.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CDO2 Limited nor the names of its contributors
 *       may be used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CDO2 LIMITED ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.scatterspace.jmx.jini;

import java.io.IOException;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.ListenerNotFoundException;
import javax.management.MBeanServerConnection;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.remote.JMXConnector;
import javax.security.auth.Subject;

import net.jini.core.lookup.ServiceTemplate;
import net.scatterspace.client.ServiceFinder;
import net.scatterspace.client.ServiceFinderResult;
import net.scatterspace.jmx.MBeanConnector;

/**
 * Implementation of the actual client connector, using the
 * {@link ServiceFinder} clientutils.
 */
public class JiniClientConnector implements JMXConnector
{
    private Logger log_ = Logger.getLogger(getClass().getName());
    
    private String[] discoveryGroups_;
    private ServiceTemplate template_;
    private long timeout_;
    private MBeanServerConnection conn_;
    
    public JiniClientConnector(String[] discoveryGroups,
                               ServiceTemplate template, long timeout)
    {
        discoveryGroups_ = discoveryGroups;
        template_ = template;
        timeout_ = timeout;
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#connect()
     */
    public void connect() throws IOException
    {
        log_.log(Level.FINE, "Connect initiated");
        
        if (conn_ != null)
            return;
        
        ServiceFinderResult finder = new ServiceFinder(
            discoveryGroups_, null).lookup(template_, null);
        
        MBeanConnector remoteConn;
        try
        {
            remoteConn = (MBeanConnector) finder.getService(timeout_);
        }
        catch (InterruptedException e)
        {
            throw new IOException("Interrupt while searching for server", e);
        }
        
        log_.log(Level.FINE, "Discovered connector service " + remoteConn);
        
        if (remoteConn == null)
            throw new IOException("No remote server found");
        
        conn_ = new ClientMBeanAdapter(remoteConn);
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#close()
     */
    public void close() throws IOException
    {
        log_.log(Level.FINE, "Connector close requested, ignoring");
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#connect(java.util.Map)
     */
    public void connect(Map<String,?> env) throws IOException
    {
        if (env != null && env.containsKey("net.scatterspace.jmx.timeout"))
            timeout_ = (Long) env.get("net.scatterspace.jmx.timeout");
        
        connect();
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#getConnectionId()
     */
    public String getConnectionId() throws IOException
    {
        log_.log(Level.FINER, "Got connection ID: " + conn_.toString());
        
        return conn_.toString();
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#getMBeanServerConnection()
     */
    public MBeanServerConnection getMBeanServerConnection() throws IOException
    {
        log_.log(Level.FINER, "Got the MBean connection");
        return conn_;
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#getMBeanServerConnection(javax.security.auth.Subject)
     */
    public MBeanServerConnection getMBeanServerConnection(Subject arg0)
            throws IOException
    {
        log_.log(Level.WARNING, "MBean server with subject requested [fail]");
        throw new UnsupportedOperationException();
    }

    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#addConnectionNotificationListener(javax.management.NotificationListener, javax.management.NotificationFilter, java.lang.Object)
     */
    public void addConnectionNotificationListener(NotificationListener arg0,
            NotificationFilter arg1, Object arg2)
    {
        log_.log(Level.WARNING, "Add conn listener requested [fail]");
        throw new UnsupportedOperationException();
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#removeConnectionNotificationListener(javax.management.NotificationListener)
     */
    public void removeConnectionNotificationListener(NotificationListener arg0)
            throws ListenerNotFoundException
    {
        log_.log(Level.WARNING, "Remove conn listener requested [fail]");
        throw new UnsupportedOperationException();        
    }
    
    /* (non-Javadoc)
     * @see javax.management.remote.JMXConnector#removeConnectionNotificationListener(javax.management.NotificationListener, javax.management.NotificationFilter, java.lang.Object)
     */
    public void removeConnectionNotificationListener(NotificationListener arg0,
            NotificationFilter arg1, Object arg2)
            throws ListenerNotFoundException
    {
        log_.log(Level.WARNING, "Remove conn listener requested [fail]");
        throw new UnsupportedOperationException();
    }
}
